import logging


logger = logging.getLogger(__name__)


class AioJsonRpcException(Exception):
    pass


class InvalidRequestException(AioJsonRpcException):
    pass


class CustomJsonRpcException(AioJsonRpcException):
    def __init__(self, code, data):
        self.code = code
        self.data = data

        if not (-32099 <= code <= -32000):
            logger.error(
                "Wrong Custom Error, custom error must be "
                "between -32099 and -32000"
            )



import inspect

from aio_jsonrpc_20.exception import InvalidRequestException


JSONRPC_VERSION = "2.0"
REQUIRED_FIELDS = set(["jsonrpc", "method"])
POSSIBLE_FIELDS = set(["jsonrpc", "method", "params", "id"])


def lazy_check_request(request):
    pass


def check_request(request):
    # check request's keys
    request_keys = set(request.keys())
    extra = request_keys - POSSIBLE_FIELDS
    missed = REQUIRED_FIELDS - request_keys
    if len(extra) > 0 or len(missed) > 0:
        msg = (
            "Invalid request."
            " Extra fields: {0},"
            " Missed fields: {1}"
        )
        raise InvalidRequestException(msg.format(extra, missed))

    # check jsonrpc's version
    if request['jsonrpc'] != JSONRPC_VERSION:
        raise InvalidRequestException(
            'version of the JSON-RPC protocol '
            'MUST be exactly "2.0"'
        )

    # check method
    method = request['method']
    if not isinstance(method, str):
        raise InvalidRequestException("Method should be string")

    if method.startswith("rpc."):
        raise InvalidRequestException(
            "Method names that begin with the word rpc followed "
            "by a period character (U+002E or ASCII 46) are "
            "reserved for rpc-internal methods and extensions "
            "and MUST NOT be used for anything else."
        )

    # check params
    if 'params' in request:
        params = request['params']
        params_check = isinstance(params, (list, tuple, dict))
        if params and not params_check:
            raise InvalidRequestException(
                "Incorrect params: {0}".format(params)
            )

    # check id
    if 'id' in request:
        request_id = request['id']
        if request_id and not isinstance(request_id, (str, int)):
            raise InvalidRequestException(
                "Incorrect id: {0}".format(request_id)
            )


def is_valid_params(func, params):
    # inspect method
    spec = inspect.getfullargspec(func)
    if spec.defaults:
        required = spec.args[:-len(spec.defaults)]
    else:
        required = spec.args

    # checks params validity
    if not params:
        if required:
            return False
    elif isinstance(params, list):
        if len(params) < len(required) or len(params) > len(spec.args):
            return False
    else:
        params_keys = set(params.keys())
        if (
                len(params_keys) > len(spec.args) or
                len(set(params_keys) & set(required)) != len(required) or
                len(set(params_keys) - set(spec.args)) > 0
        ):
            return False

    return True



import json

from aio_jsonrpc_20.utils import check_request, lazy_check_request


class RequestBuilder():
    def __init__(self, serializer=json, lazy_check=False):
        self.serializer = serializer
        self.inc = 1

        if lazy_check:
            self.check_request = lazy_check_request
        else:
            self.check_request = check_request

    def call(self, method, params=None):
        request = {
            "jsonrpc": "2.0",
            "method": method,
            "id": self.inc
        }

        if params:
            request["params"] = params

        self.check_request(request)

        self.inc += 1

        if self.serializer:
            return self.serializer.dumps(request)
        else:
            return request

    def notify(self, method, params=None):
        request = {
            "jsonrpc": "2.0",
            "method": method
        }

        if params:
            request["params"] = params

        self.check_request(request)

        if self.serializer:
            return self.serializer.dumps(request)
        else:
            return request


class BatchRequestBuilder():
    BUILDER_CLASS = RequestBuilder

    def __init__(self, serializer=json):
        self.serializer = serializer
        self.builder = self.BUILDER_CLASS(serializer=None)
        self.batch = []

    def call(self, method, params=None):
        request = self.builder.call(method, params)
        self.batch.append(request)

        return request["id"]

    def notify(self, method, params=None):
        request = self.builder.notify(method, params)
        self.batch.append(request)

    def get_request(self):
        if self.serializer:
            return self.serializer.dumps(self.batch)
        else:
            return self.batch

    def purge(self):
        self.batch = []





class ResponseMaker(object):
    __slot__ = ['error_verbose']

    def __init__(self, error_verbose=True):
        self.error_verbose = error_verbose

    def get_response(self, result, request_id):
        return {
            "jsonrpc": "2.0",
            "result": result,
            "id": request_id
        }

    def get_error(self, code, message, data=None, request_id=None):
        result = {
            "jsonrpc": "2.0",
            "error": {
                "code": code,
                "message": message,
            },
            "id": request_id
        }

        if self.error_verbose and data:
            result["error"]['data'] = data

        return result

    def get_parse_error(self, data=None, request_id=None):
        return self.get_error(
            -32700, 'Parse error', data=data, request_id=request_id
        )

    def get_invalid_request(self, data=None, request_id=None):
        return self.get_error(
            -32600, 'Invalid Request', data=data, request_id=request_id
        )

    def get_method_not_found(self, data=None, request_id=None):
        return self.get_error(
            -32601, 'Method not found', data=data, request_id=request_id
        )

    def get_invalid_params(self, data=None, request_id=None):
        return self.get_error(
            -32602, 'Invalid params', data=data, request_id=request_id
        )

    def get_internal_error(self, data=None, request_id=None):
        return self.get_error(
            -32603, 'Internal error', data=data, request_id=request_id
        )

    def get_server_error(self, code, data, request_id=None):
        return self.get_error(
            code=code,
            message='Server error',
            data=data,
            request_id=request_id
        )



import asyncio
import json
import logging

from aio_jsonrpc_20.exception import (
    InvalidRequestException,
    CustomJsonRpcException
)
from aio_jsonrpc_20.response import ResponseMaker
from aio_jsonrpc_20.utils import (
    is_valid_params,
    check_request,
    lazy_check_request
)


logger = logging.getLogger(__name__)


class RequestResolver(object):
    """This class allow to resolve a request, and build a response."""
    __slots__ = [
        'router',
        'serializer',
        'response_maker',
        'check_request'
    ]

    def __init__(
            self,
            router,
            lazy_check=False,
            error_verbose=True,
            serializer=json
    ):
        self.router = router
        self.response_maker = ResponseMaker(error_verbose)
        self.serializer = serializer

        if lazy_check:
            self.check_request = lazy_check_request
        else:
            self.check_request = check_request

    async def handle(self, str_request):
        # handle encoding
        if isinstance(str_request, bytes):
            str_request = str_request.decode("utf-8")

        # get response from unserialized request
        try:
            request = self.serializer.loads(str_request)
        except (TypeError, ValueError):
            logger.warning('Bad formatted json')
            response = self.response_maker.get_parse_error(
                data='Bad formatted json'
            )
        else:
            if request:
                try:
                    # batch request case
                    if isinstance(request, list):
                        coros = []
                        for sub_req in request:
                            coros.append(self._get_response(sub_req))
                        response = await asyncio.gather(*coros)
                        response = [elt for elt in response if elt]
                    # simple request case
                    else:
                        response = await self._get_response(request)
                # handle uncaught exception
                except Exception as e:
                    logger.warning('Unexpected error: ' + str(e))
                    response = self.response_maker.get_internal_error(
                        data=e.args[0]
                    )
            else:
                response = self.response_maker.get_invalid_request(
                    data='Empty request is not allowed'
                )

        # return serialized result
        return self.serializer.dumps(response) if response else ''

    async def _get_response(self, request):
        # get id and check in the same time if request is a dict
        try:
            request_id = request['id']
        except KeyError:
            request_id = None
        except TypeError:
            logger.warning('Request should be a dict object')
            return self.response_maker.get_invalid_request(
                data="Request should be a dict object",
            )

        # check jsonrpc 2.0 specification
        try:
            self.check_request(request)
        except InvalidRequestException as e:
            logger.warning('Invalid request')
            return self.response_maker.get_invalid_request(
                data=e.args[0],
                request_id=request_id
            )

        # get method to call
        try:
            method = self.router[request['method']]
        except KeyError:
            logger.warning('Method not found')
            return self.response_maker.get_method_not_found(
                request_id=request_id
            )
        except TypeError:
            logger.error('Router should be like an dict object')
            return self.response_maker.get_internal_error(
                data="Router should be like an dict object",
            )

        # try to execute method
        try:
            if 'params' in request:
                params = request['params']
                if isinstance(params, dict):
                    result = await method(**params)
                else:
                    result = await method(*params)
            else:
                result = await method()
        except CustomJsonRpcException as e:
            # catch custom error
            logger.warning('CustomError: ' + str(e))
            return self.response_maker.get_server_error(
                code=e.code,
                data=e.data,
                request_id=request_id
            )
        except Exception as e:
            # determine the error's type
            if (
                    isinstance(e, TypeError) and
                    not is_valid_params(method, params)
            ):
                logger.warning('Invalid params')
                return self.response_maker.get_invalid_params(
                    data=e.args[0], request_id=request_id
                )
            else:
                logger.warning('Application error: ' + str(e))
                return self.response_maker.get_internal_error(
                    data=e.args[0], request_id=request_id
                )

        # is notification ?
        if request_id:
            return self.response_maker.get_response(
                result=result, request_id=request_id
            )
        else:
            return None



from aio_jsonrpc_20.builder import RequestBuilder, BatchRequestBuilder
from aio_jsonrpc_20.exception import CustomJsonRpcException
from aio_jsonrpc_20.resolver import RequestResolver
from aio_jsonrpc_20.version import __version__

__all__ = [
    'BatchRequestBuilder',
    'RequestBuilder',
    'RequestResolver',
    'CustomJsonRpcException',
    '__version__'
]



#!/usr/bin/env python
import os

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
__version__ = None
with open('aio_jsonrpc_20/version.py') as f:
    exec(f.read())

with open(os.path.join(here, 'README.md')) as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.md')) as f:
    CHANGES = f.read()

desc = "fast JSON-RPC protocol implementation for asyncio, without transport"

setup(
    name="aio-jsonrpc-2.0",
    version=__version__,
    packages=find_packages(),
    # metadata for upload to PyPI
    author="Lahache Stephane",
    author_email='slahache@gmail.com',
    url="https://github.com/steffgrez/aio-jsonrpc-2.0",
    description=desc,
    long_description=README + '\n\n' + CHANGES,
    keywords='json rpc jsonrpc json-rpc 2.0',

    # Full list:
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        'Intended Audience :: Developers',
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    license="MIT",
)
